module Game where

    import Graphics.Gloss
    import Graphics.Gloss.Interface.Pure.Game
    import Window
    import Pacman
    import Collision
    import Ghost
    import Coins

    data GameStatus = Game
        { playerLoc  :: Position        -- ^ Posição do pacman 
        , playerVelX :: Float           -- ^ Velocidade do pacman na horizontal x
        , playerVelY :: Float           -- ^ Velocidade do pacman na vertical y
        , playerAcc  :: Float           -- ^ Aceleração do jogador.
        , ghostLoc   :: Position        -- ^ Posição do fantasma
        , ghostVelX  :: Float           -- ^ Velocidade do fantasma na horizontal x
        , ghostVelY  :: Float           -- ^ Velocidade do fantasma na vertical y
        , wallsList  :: [Window.Wall]   -- ^ Lista de paredes do mapa
        , ghostAcc   :: Float           -- ^ Aceleração do fantasma
        , gameStat   :: Int             -- ^ status do jogo
        , coins      :: Coins           -- ^ lista de moeda
        , isPaused   :: Bool            -- ^ Verifica se o jogo está pausado
        }

    -- | Estado inicial do jogo
    initialState :: GameStatus
    initialState = Game
        { playerLoc  = (0,0)
        , playerVelX = 0
        , playerVelY = 0
        , playerAcc  = 100
        , ghostLoc   = (170, 170) -- fantasma inicia no canto direito pra cima do mapa
        , ghostVelX  = 0
        , ghostVelY  = 0
        , wallsList  = wallsArray -- chama o vetor criado no Window.hs
        , ghostAcc   = 70
        , gameStat   = 0
        , coins      = genCoins -- chama a função que gera a lista das moedas
        , isPaused   = True -- Jogo começa pausado
        }

    -- | Converte o estado do jogo em uma imagem de tela.
    render :: GameStatus -> Picture
    render game = pictures [walls, playerPic, ghostPic, msgPic, coinsPic]
        where
            playerPic = mkPlayer   (playerLoc game)
            ghostPic  = mkGhost    (ghostLoc game)
            coinsPic  = drawCoins  (coins game)
            msgPic    = curMsg     (gameStat game) (isPaused game)


    -- As funções a seguir estavam no jogo Arcanoid dado no curso de Haskell http://pesquisa.ufabc.edu.br/haskell/posts/bibliotecas/01-Gloss.html

    -- | Incrementa a velocidade do jogador.
    incVelX :: GameStatus -> GameStatus
    incVelX game = game { playerVelX = playerAcc game, playerVelY = 0 }

    -- | Incrementa a velocidade do jogador.
    incVelY :: GameStatus -> GameStatus
    incVelY game = game { playerVelY = playerAcc game, playerVelX = 0 }

    -- | Decrementa a velocidade do jogador.
    decVelX :: GameStatus -> GameStatus
    decVelX game = game { playerVelX = -(playerAcc game), playerVelY = 0 }

    -- | Decrementa a velocidade do jogador.
    decVelY :: GameStatus -> GameStatus
    decVelY game = game { playerVelY = -(playerAcc game), playerVelX = 0  }

    -- | Inverte o estado de pausa do jogo.
    invPause :: GameStatus -> GameStatus
    invPause game = game { isPaused = not $ isPaused game }

    -- | Responde aos e ventos de teclas.
    handleKeys :: Event -> GameStatus -> GameStatus
    -- Tecla 'r' retorna ao estaod inicial.
    handleKeys (EventKey (Char 'r') Down _ _)            game = initialState
    -- Tecla 'p' pausa e despausa o jogo.
    handleKeys (EventKey (Char 'p') Down _ _)            game = invPause game
    -- Tecla '⬅️' move para esquerda.
    handleKeys (EventKey (SpecialKey KeyLeft) Down _ _)  game = decVelX game
    -- Tecla '➡️' move o jogador para a direita.
    handleKeys (EventKey (SpecialKey KeyRight) Down _ _) game = incVelX game
    -- Tecla '⬆️' move o jogador para cima.
    handleKeys (EventKey (SpecialKey KeyUp) Down _ _)  game = incVelY game
    -- Tecla '⬇️' move o jogador para baixo.
    handleKeys (EventKey (SpecialKey KeyDown) Down _ _)  game = decVelY game

    -- Qualquer outra tecla é ignorada.
    handleKeys _ game = game

    -- | Verifica a posição do jogagor
    updatePlayer :: Float -> GameStatus -> GameStatus
    updatePlayer seconds game = game { playerLoc = movePacman seconds x (vx, vy) }
        where 
            x        = playerLoc game
            vx       = playerVelX game
            vy       = playerVelY game 

    -- | Verifica a posição do Pacman de acordo com as colisões
    updateWallPlayer :: GameStatus -> GameStatus
    updateWallPlayer game = game { playerVelX = vx, playerVelY = vy }
        where 
            pos      = playerLoc game
            v        = (playerVelX game, playerVelY game)
            (vx, vy) = wallBounce pos v

    -- | Fantasma que segue o Pacman horizontalmente preferencialmente
    updateGhost :: Float -> GameStatus -> GameStatus
    updateGhost seconds game  | leftCollisionGhost x && downCollision x   = game { ghostLoc = moveGhost seconds x (ghoAcc, 0) } -- canto esquerdo pra baixo
                              | leftCollisionGhost x && topCollision x   = game { ghostLoc = moveGhost seconds x (0, (-ghoAcc)) } -- canto esquerdo pra cima
                              | rightCollisionGhost x && downCollision x  = game { ghostLoc = moveGhost seconds x (0, ghoAcc) } -- canto direito pra baixo
                              | rightCollisionGhost x && topCollision x  = game { ghostLoc = moveGhost seconds x ((-ghoAcc), 0) } -- canto direito pra cima 
                              | wallCollision11 x                     = game { ghostLoc = moveGhost seconds x (0, ghoAcc)} 
                              | wallCollision10 x                     = game { ghostLoc = moveGhost seconds x (0,(-ghoAcc))}
                              | wallCollision3 x && wallCollision7 x  = game { ghostLoc = moveGhost seconds x (0, (-ghoAcc))}
                              | wallCollision5 x && wallCollision9 x  = game { ghostLoc = moveGhost seconds x (ghoAcc, 0)}
                              | wallCollision4 x && wallCollision8 x  = game { ghostLoc = moveGhost seconds x (0, (ghoAcc))}
                              | wallCollision6 x && wallCollision2 x  = game { ghostLoc = moveGhost seconds x ((-ghoAcc), 0)}
                              | wallCollision x && ghoVx < pacVx      = game { ghostLoc = moveGhost seconds x (ghoAcc,0)}
                              | wallCollision x && ghoVx >= pacVx     = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)}
                              | wallCollision1 x && ghoVx < pacVx     = game { ghostLoc = moveGhost seconds x (ghoAcc,0)}
                              | wallCollision1 x && ghoVx >= pacVx     = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)}
                              | wallCollision6 x && ghoVx < pacVx     = game { ghostLoc = moveGhost seconds x (ghoAcc,0)}
                              | wallCollision6 x && ghoVx >= pacVx     = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)}
                              | wallCollision7 x && ghoVx < pacVx     = game { ghostLoc = moveGhost seconds x (ghoAcc,0)}
                              | wallCollision7 x && ghoVx >= pacVx     = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)}
                              | wallCollision8 x && ghoVx < pacVx     = game { ghostLoc = moveGhost seconds x (ghoAcc,0)}
                              | wallCollision8 x && ghoVx >= pacVx     = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)}
                              | wallCollision9 x && ghoVx < pacVx     = game { ghostLoc = moveGhost seconds x (ghoAcc,0)}
                              | wallCollision9 x && ghoVx >= pacVx     = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)}
                              | wallCollision2 x || wallCollision4 x  = game { ghostLoc = moveGhost seconds x (0, ghoAcc)}
                              | wallCollision3 x || wallCollision5 x  = game { ghostLoc = moveGhost seconds x (0, (-ghoAcc))}
                              | downCollision x                       = game { ghostLoc = moveGhost seconds x (ghoAcc,0) } -- parede de baixo
                              | topCollision x                        = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0) } -- parede de cima
                              | rightCollisionGhost x                 = game { ghostLoc = moveGhost seconds x (0, ghoAcc) } -- parede da direita
                              | leftCollisionGhost x                  = game { ghostLoc = moveGhost seconds x (0, (-ghoAcc)) } -- parede da esquerda
                            --   | abs (ghoVx-pacVx) < 20 && ghoVy>pacVy = game { ghostLoc = moveGhost seconds x (0, (-ghoAcc)) } -- verifica a faixa de valor em x do fantasma e do pacman, se estiver dentro do valor limitado, se o pacman estiver para CIMA ele ir para CIMA
                            --   | abs (ghoVx-pacVx) < 20 && ghoVy<pacVy = game { ghostLoc = moveGhost seconds x (0, ghoAcc) } -- verifica a faixa de valor em x do fantasma e do pacman, se estiver dentro do valor limitado, se o pacman estiver para BAIXO ele ir para BAIXO
                              | ghoVx < pacVx                         = game { ghostLoc = moveGhost seconds x (ghoAcc,0) } -- se não estiver encostado em nenhuma parede, ele verifica se o pacman está para DIREITA, para ir para a DIREITA
                              | ghoVx >= pacVx                        = game { ghostLoc = moveGhost seconds x ((-ghoAcc),0)} -- se não estiver encostado em nenhuma parede, ele verifica se o pacman está para ESQUERDA, para ir para a ESQUERDA
                              | otherwise                             = game { ghostLoc = moveGhost seconds x (0, ghoAcc) }

        where 
            x               = ghostLoc game
            v               = (ghostVelX game, ghostVelY game)
            (pacVx, pacVy)  = playerLoc game
            (ghoVx, ghoVy)  = ghostLoc game
            ghoAcc          = ghostAcc game
            leftCollisionG  = leftCollisionGhost (ghostLoc game)
            rightCollisionG = rightCollisionGhost (ghostLoc game)

    -- | Verifica se as moedas foram coletadas
    updateCoins :: GameStatus -> GameStatus
    updateCoins game | coletouCoin pos cs = game {coins = coins'}
                     | otherwise = game
        where
            pos = playerLoc game
            cs  = coins game
            coins' = removeCoin pos (coins game)
    
    -- | Atualiza o estado do jogo.
    update :: Float -> GameStatus -> GameStatus
    update seconds game | isPaused game               = game
                        | (not.hasCoins) $ coins game = game { gameStat = 1 }
                        | dropped                     = game { gameStat = -1 }
                        | otherwise                   = collisions $ moves game
        where
            dropped    = gameOver (playerLoc game) (ghostLoc game)
            collisions = updateWallPlayer . updateCoins
            moves      = updatePlayer seconds . updateGhost seconds
